from os import path as osp

# from .basic import Meme
# from .thug import ThugMeme

DATA_FOLDER = osp.join(osp.dirname(__file__), 'data')
GLASSES_FILE = osp.join(DATA_FOLDER, 'glasses.png')
CIGAR_FILE = osp.join(DATA_FOLDER, 'cigar.png')
FONT = osp.join(DATA_FOLDER, 'font.ttf')



from enum import Enum
from os import (getcwd, path as osp)
import sys

import click
import cv2

from .conf import (get_config, print_config)
from .detect.opencv import HaarCascadeDetector
from .meme.basic import Meme
from .meme.thug import ThugMeme

MEME_RESULT_DIR = getcwd()

CONTEXT = dict(help_option_names=['-h', '--help'])


class Detector(Enum):
    OPEN_CV = 'opencv'
    DLIB = 'dlib'


_common_decorators = [
    click.version_option(None, '-v', '--version'),
    click.argument('fpath', type=click.Path(exists=True)),
    click.argument('txt1'),
    click.argument('txt2'),
    click.option(
        '--override',
        '-o',
        type=(str, str),
        multiple=True,
        help='Override any configuration option: <option_name> <new_value>.'),
    click.option(
        '--show-config',
        is_flag=True,
        help='Show the configuration and exit. Takes into account -o options.')
]


def add_decorators(decorators):
    def _add_decorators(func):
        for decorator in reversed(decorators):
            func = decorator(func)
        return func

    return _add_decorators


def _load_configuration(override, show_and_exit):
    conf = get_config(overrides=override)
    if show_and_exit:
        print_config(conf)
        sys.exit(0)

    return conf


def _form_result_path(orig_path, result_dir, fname_extra=''):
    fname = osp.basename(orig_path)
    base, extension = fname.split('.')
    fname = '{}{}.{}'.format(base, fname_extra, extension)
    return osp.join(result_dir, fname)


@click.command(context_settings=CONTEXT)
@add_decorators(_common_decorators)
def meme(fpath, txt1, txt2, override, show_config):
    """Generate a normal meme."""

    conf = _load_configuration(override, show_config)
    res_path = _form_result_path(
        orig_path=osp.abspath(fpath),
        result_dir=MEME_RESULT_DIR,
        fname_extra=conf['meme']['meme_result_name_add'])
    meme = Meme(config=conf['meme'], img_path=fpath, txt1=txt1, txt2=txt2)
    meme.create(res_file=res_path)


@click.command(context_settings=CONTEXT)
@add_decorators(_common_decorators)
@click.option(
    '--debug',
    is_flag=True,
    help='Show debug information (e.g. the detection results img)')
@click.option(
    '--detector',
    type=click.Choice([Detector.OPEN_CV.value, Detector.DLIB.value]),
    default=Detector.OPEN_CV.value,
    help='Detector to use for finding faces and landmarks.')
def thug_meme(fpath, txt1, txt2, override, show_config, debug, detector):
    """Generate an awesome thug meme."""
    fpath = osp.abspath(fpath)
    conf = _load_configuration(override, show_config)
    res_path = _form_result_path(
        orig_path=fpath,
        result_dir=MEME_RESULT_DIR,
        fname_extra=conf['meme']['thug_result_name_add'])

    if detector == Detector.OPEN_CV.value:
        detector = HaarCascadeDetector(config=conf['detect'])
    elif detector == Detector.DLIB.value:
        from .detect.dlib import DlibDetector
        detector = DlibDetector(config=conf['detect'])

    thugs = detector.find_thug_landmarks(
        img_path=osp.abspath(fpath), show_result=debug)

    meme = ThugMeme(
        config=conf['meme'],
        thug_landmarks=thugs,
        img_path=fpath,
        txt1=txt1,
        txt2=txt2)
    meme.create(res_path)

    if debug:
        cv2.waitKey(0)
        cv2.destroyAllWindows()



from os import path as osp


def img_path(fname):
    return osp.abspath(osp.join(IMG_DIR, fname))


IMG_DIR = osp.join(osp.dirname(__file__), 'data', 'img')
IMG_1_FACE = img_path('1_face.jpg')
IMG_3_FACES = img_path('3_faces.jpeg')



import cv2
import dlib
import face_recognition_models

from .abc import Detector
from .debug import timestamp  # TODO: get rid off this
from .result import ThugLandmarks

LEFT_EYE_CENTER_IDXS = (37, 40)
RIGHT_EYE_CENTER_IDXS = (43, 46)
MOUTH_CENTER_IDXS = (62, 66)


class DlibDetector(Detector):
    def __init__(self, config, *args, **kwargs):
        self._landmarks_model = face_recognition_models.pose_predictor_model_location(
        )
        super().__init__(config, *args, **kwargs)

    def find_thug_landmarks(self, img_path, show_result=False):
        face_det = dlib.get_frontal_face_detector()

        predictor = dlib.shape_predictor(self._landmarks_model)

        img = cv2.imread(img_path)
        gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

        faces = face_det(gray)

        thug_results = []

        for f in faces:
            predicted = predictor(gray, f)
            landmarks = predicted.parts()

            left_eye = self._center_of_two_points(
                landmarks[LEFT_EYE_CENTER_IDXS[0]],
                landmarks[LEFT_EYE_CENTER_IDXS[1]])

            right_eye = self._center_of_two_points(
                landmarks[RIGHT_EYE_CENTER_IDXS[0]],
                landmarks[RIGHT_EYE_CENTER_IDXS[1]])

            mouth = self._center_of_two_points(landmarks[MOUTH_CENTER_IDXS[0]],
                                               landmarks[MOUTH_CENTER_IDXS[1]])

            thug = ThugLandmarks(l_eye=left_eye, r_eye=right_eye, mouth=mouth)
            thug_results.append(thug)

            if show_result:
                self._draw_result(img, f, landmarks, thug)

        if show_result:
            cv2.imshow('detection_result-{}'.format(timestamp()), img)
            cv2.waitKey(1)

        return thug_results

    def _draw_result(self, img, face, landmarks, thug_result):
        for point in landmarks:
            self._draw_circle(img, (point.x, point.y), bgr=(0, 255, 255))
        self._draw_thug_result(img, thug_result)
        self._draw_face(img, self.dlib_rect_to_xywh(face))

    @staticmethod
    def _center_of_two_points(point1, point2):
        x = int(point1.x + (point2.x - point1.x) / 2)
        y = int(point1.y + (point2.y - point1.y) / 2)
        return x, y

    @staticmethod
    def dlib_rect_to_xywh(r):
        x, y = r.left(), r.top()
        w, h = r.right() - x, r.bottom() - y
        return x, y, w, h



from unittest.mock import patch

import pytest

from thug.meme.basic import Meme
from thug.conf import get_config
from ..conftest import IMG_1_FACE


@pytest.fixture
def meme():
    conf = get_config()['meme']  # Default config
    m = Meme(
        config=conf, img_path=IMG_1_FACE, txt1='upper text', txt2='lower text')
    return m


@patch('thug.meme.basic.cv2.waitKey')
@patch('thug.meme.basic.cv2.imshow')
@patch('thug.meme.basic.cv2.imwrite')
class TestCreate:
    @patch('thug.meme.basic.cv2.putText')
    def test_it_writes_texts(self, puttext, imwrite, imshow, waitkey, meme):
        meme.create(res_file='test', show=False)

        assert puttext.call_count == 2
        texts = [args[1] for (args, _) in puttext.call_args_list]
        assert texts[0] == 'upper text'
        assert texts[1] == 'lower text'
        assert not imshow.call_count

    def test_it_shows_result(self, imwrite, imshow, waitkey, meme):
        meme.create(res_file='test', show=True)
        assert imshow.call_count == 1



"""Common tests for all detectors"""

from collections import namedtuple
from unittest.mock import patch

import pytest

from thug.conf import get_config
from thug.detect.opencv import HaarCascadeDetector
from thug.detect.dlib import DlibDetector
from ..conftest import (IMG_1_FACE, IMG_3_FACES)

Tolerance = namedtuple('Tolerance', 'min max')
_TestCase = namedtuple('TestCase', 'input_file tolerances')


class LocationTolerances:
    def __init__(self,
                 left_eye_x=None,
                 left_eye_y=None,
                 right_eye_x=None,
                 right_eye_y=None,
                 mouth_x=None,
                 mouth_y=None):
        self.left_eye_x = left_eye_x
        self.left_eye_y = left_eye_y
        self.right_eye_x = right_eye_x
        self.right_eye_y = right_eye_y
        self.mouth_x = mouth_x
        self.mouth_y = mouth_y


def _test_cases():
    TOLERANCES_1 = [
        LocationTolerances(
            left_eye_x=Tolerance(188, 195),
            left_eye_y=Tolerance(320, 327),
            right_eye_x=Tolerance(282, 289),
            right_eye_y=Tolerance(299, 305),
            mouth_x=Tolerance(270, 282),
            mouth_y=Tolerance(420, 430))
    ]

    # Ordering left to right
    TOLERANCES_2 = [
        LocationTolerances(
            left_eye_x=Tolerance(385, 395),
            left_eye_y=Tolerance(346, 356),
            right_eye_x=Tolerance(436, 446),
            right_eye_y=Tolerance(350, 360),
            mouth_x=Tolerance(405, 415),
            mouth_y=Tolerance(400, 410)),
        LocationTolerances(
            mouth_x=Tolerance(872, 882), mouth_y=Tolerance(282, 292)),
        LocationTolerances(
            mouth_x=Tolerance(1227, 1237), mouth_y=Tolerance(455, 465))
    ]

    case1 = _TestCase(input_file=IMG_1_FACE, tolerances=TOLERANCES_1)
    case2 = _TestCase(input_file=IMG_3_FACES, tolerances=TOLERANCES_2)

    for case in [case1, case2]:
        yield case


class TestFindThugLandMarks:
    @pytest.mark.parametrize('test_case', _test_cases())
    @pytest.mark.parametrize('detector', [HaarCascadeDetector, DlibDetector])
    def test_it_returns_landmarks_within_tolerances(self, detector, test_case):
        conf = get_config()['detect']  # Default config
        d = detector(config=conf)
        res = d.find_thug_landmarks(
            img_path=test_case.input_file, show_result=False)

        # verify face count
        assert len(res) == len(test_case.tolerances)

        # sort from left to right
        sorted_res = sorted(res, key=lambda l: l.left_eye.x)

        for result, t in zip(sorted_res, test_case.tolerances):
            if t.mouth_x:
                assert t.mouth_x.min <= result.mouth.x <= t.mouth_x.max
            if t.mouth_y:
                assert t.mouth_y.min <= result.mouth.y <= t.mouth_y.max

            if t.left_eye_x:
                assert t.left_eye_x.min <= result.left_eye.x <= t.left_eye_x.max
            if t.left_eye_y:
                assert t.left_eye_y.min <= result.left_eye.y <= t.left_eye_y.max

            if t.right_eye_x:
                assert t.right_eye_x.min <= result.right_eye.x <= t.right_eye_x.max
            if t.left_eye_y:
                assert t.right_eye_y.min <= result.right_eye.y <= t.right_eye_y.max

    @pytest.mark.parametrize(
        'detector, imshow_path',
        [(HaarCascadeDetector, 'thug.detect.opencv.cv2.imshow'),
         (DlibDetector, 'thug.detect.dlib.cv2.imshow')])
    def test_it_shows_results(self, detector, imshow_path):
        conf = get_config()['detect']
        d = detector(config=conf)
        with patch(imshow_path) as imshow:
            d.find_thug_landmarks(img_path=IMG_1_FACE, show_result=True)
            assert imshow.call_count



from abc import (ABC, abstractmethod)

import cv2


class Detector(ABC):
    def __init__(self, config, *args, **kwargs):
        self._config = config
        pass

    @property
    def config(self):
        return self._config

    @abstractmethod
    def find_thug_landmarks(self, img_path, show_result=False):
        """
        returns a list of ThugLandmark objects
        """

    @staticmethod
    def _draw_rectangle(img, rect, bgr, thickness=2):
        x, y, w, h = rect
        cv2.rectangle(img, (x, y), (x + w, y + h), bgr, thickness)

    @staticmethod
    def _draw_circle(img, center, bgr, radius=5, thickness=2):
        cv2.circle(img, center, radius, bgr, thickness)

    def _draw_face(self, img, location):
        bgr = (255, 0, 0)
        self._draw_rectangle(img, location, bgr)

    def _draw_thug_result(self, img, thug_result):
        bgr_eye = (0, 0, 255)
        bgr_mouth = (0, 0, 255)
        for eye in (thug_result.left_eye, thug_result.right_eye):
            self._draw_circle(img, eye, bgr_eye, thickness=5)
        self._draw_circle(img, thug_result.mouth, bgr_mouth, thickness=5)


from unittest.mock import (patch, MagicMock)

import cv2
import pytest

from thug.detect.result import ThugLandmarks
from thug.meme.thug import ThugMeme
from thug.conf import get_config
from ..conftest import IMG_3_FACES

DEFAULT_KWARGS = {
    'config': get_config()['meme'],
    'txt1': 'upper txt',
    'txt2': 'lower txt',
    'img_path': IMG_3_FACES
}


@pytest.fixture
def thug1():
    thug = ThugLandmarks(l_eye=(100, 100), r_eye=(150, 100), mouth=(125, 300))
    return thug


@pytest.fixture
def thug2():
    thug = ThugLandmarks(l_eye=(250, 125), r_eye=(300, 125), mouth=(275, 350))
    return thug


@pytest.fixture
def thug_meme(thug1, thug2):
    meme = ThugMeme(thug_landmarks=[thug1, thug2], **DEFAULT_KWARGS)
    return meme


class TestCreate:
    @patch('thug.meme.basic.cv2.imwrite')
    @patch('thug.meme.basic.Meme.create')
    @patch('thug.meme.thug.ThugMeme._draw_glasses')
    @patch('thug.meme.thug.ThugMeme._draw_cigar')
    def test_it_draws_cigar_and_glasses_for_all_thugs(
            self, draw_cigar, draw_glasses, base_cls_create, imwrite,
            thug_meme):

        thug_meme.create(res_file='test', show=False)

        assert draw_glasses.call_count == 2
        assert draw_cigar.call_count == 2
        assert base_cls_create.call_count == 1


class TestDrawCigar:
    @patch('thug.meme.thug.ThugMeme._draw_on_top')
    def test_it_draws(self, draw_on_top, thug_meme, thug1):
        img = cv2.imread(IMG_3_FACES)
        thug_meme._draw_cigar(img, thug1)

        assert draw_on_top.call_count == 1
        args, _ = draw_on_top.call_args
        assert (args[0] == img).all()


class TestDrawGlasses:
    @patch('thug.meme.thug.ThugMeme._draw_on_top')
    def test_it_draws(self, draw_on_top, thug_meme, thug1):
        img = cv2.imread(IMG_3_FACES)
        thug_meme._draw_glasses(img, thug1)

        assert draw_on_top.call_count == 1
        args, _ = draw_on_top.call_args
        assert (args[0] == img).all()



from os import path as osp

DATA_FOLDER = osp.join(osp.dirname(__file__), 'data')
FACE_CASCADE_FILE = osp.join(DATA_FOLDER,
                             'haarcascade_frontalface_default.xml')
EYE_CASCADE_FILE = osp.join(DATA_FOLDER, 'haarcascade_eye.xml')
MOUTH_CASCADE_FILE = osp.join(DATA_FOLDER, 'haarcascade_mouth.xml')
DLIB_68_LANDMARKS_MODEL_FILE = osp.join(
    DATA_FOLDER, 'shape_predictor_68_face_landmarks.dat')



import json

import cv2
import datetime as dt
import numpy as np
import random

from .basic import Meme
from . import (GLASSES_FILE, CIGAR_FILE)


class ThugMeme(Meme):
    def __init__(self, *, thug_landmarks, config, **kwargs):
        self._glasses_width = float(config['glasses_width'])
        self._glasses_down_shift = float(config['glasses_down_shift'])
        self._glasses_right_shift = float(config['glasses_right_shift'])
        self._cigar_length = float(config['cigar_length'])
        self._cigar_mouth_center_pos = tuple(
            json.loads(config['cigar_mouth_center_pos']))
        self._thugs = thug_landmarks
        super().__init__(config=config, **kwargs)

    def create(self, res_file, show=True):
        img = cv2.imread(self._img_path)

        for thug in self._thugs:
            if thug.eyes_available:
                self._draw_glasses(img, thug)

                if thug.mouth_available:
                    self._draw_cigar(img, thug)  # depends also on eyes

        cv2.imwrite(res_file, img)
        self._img_path = res_file

        return super().create(res_file, show)

    def _draw_glasses(self, img, thug):
        orig_img = cv2.imread(GLASSES_FILE, -1)
        rotated = self._rotate(orig_img, thug.eyes_angle)
        h, w, _ = rotated.shape

        gx, gy, gw, gh = self._calculate_glasses_dimensions(w, h, thug)
        scaled = cv2.resize(rotated, (gw, gh), interpolation=cv2.INTER_AREA)

        self._draw_on_top(img, gx, gy, scaled)

    def _calculate_glasses_dimensions(self, origw, origh, thug):
        w = int(thug.eyes_distance * self._glasses_width)
        h = int(origh * (w / origw))

        center_x, center_y = thug.eyes_center
        y = int(center_y - h / 2 + (self._glasses_down_shift * h))
        x = int(center_x - w / 2 + (self._glasses_right_shift * w))

        return x, y, w, h

    def _rotate(self, img, angle):
        rot_matrix, w, h = self._get_rotation_matrix(img, angle)
        rot_img = cv2.warpAffine(img, rot_matrix, (w, h))
        return rot_img

    def _get_rotation_matrix(self, img, angle):
        orig_h, orig_w, _ = img.shape
        c_x, c_y = orig_w // 2, orig_h // 2

        rot_matrix = cv2.getRotationMatrix2D((c_x, c_y), -angle, 1.0)
        cos = np.abs(rot_matrix[0, 0])
        sin = np.abs(rot_matrix[0, 1])

        w = int((orig_h * sin) + (orig_w * cos))
        h = int((orig_h * cos) + (orig_w * sin))

        rot_matrix[0, 2] += (w / 2) - c_x
        rot_matrix[1, 2] += (h / 2) - c_y

        return rot_matrix, w, h

    def _get_rotated_point(self, img, angle, point):
        rot_matrix, _, _ = self._get_rotation_matrix(img, angle)
        rot_point = np.dot(rot_matrix, [point[0], point[1], 1])
        rot_point = int(rot_point[0]), int(rot_point[1])
        return rot_point

    def _draw_cigar(self, img, thug):
        def rand_angle():
            random.seed(dt.datetime.now())
            return random.randrange(360)

        orig = cv2.imread(CIGAR_FILE, -1)
        orig_h, orig_w, _ = orig.shape
        angle = rand_angle()
        rotated = self._rotate(orig, angle)

        rot_h, rot_w, _ = rotated.shape
        w, h = self._calculate_cigar_wh(rot_w, rot_h, thug)
        scaled_rot = cv2.resize(rotated, (w, h))

        scaled_orig_h = int(orig_h * (h / rot_h))
        scaled_orig_w = int(orig_w * (w / rot_w))
        scaled_orig = cv2.resize(orig, (scaled_orig_w, scaled_orig_h))
        cigar_mouth_center = self._get_rotated_point(
            scaled_orig, angle, self._cigar_mouth_center_pos)
        x = thug.mouth.x - cigar_mouth_center[0]
        y = thug.mouth.y - cigar_mouth_center[1]

        self._draw_on_top(img, x, y, scaled_rot)

    def _calculate_cigar_wh(self, orig_w, orig_h, thug):
        h = int(self._cigar_length * (thug.mouth.y - thug.left_eye.y))
        w = int(orig_w * (h / orig_h))
        return w, h

    def _draw_on_top(self, img, x, y, sub_img):
        # TODO: some intelligent scaling needed here if sub img does not fit
        h, w, _ = sub_img.shape
        mask = sub_img[:, :, 3]
        mask_inv = cv2.bitwise_not(mask)
        sub_img_ = sub_img[:, :, :3]

        background = img[y:y + h, x:x + w]
        background = cv2.bitwise_and(background, background, mask=mask_inv)
        foreground = cv2.bitwise_and(sub_img_, sub_img_, mask=mask)
        sum_ = cv2.add(background, foreground)

        img[y:y + h, x:x + w] = sum_



from configparser import ConfigParser
import logging
from os import (path as osp, environ as osenv)

DEFAULT_CONFIG = osp.join(osp.dirname(__file__), 'default.conf')
LOGGER = logging.getLogger(__name__)


def print_config(conf):
    from pprint import pprint
    for section in conf.sections():
        section_format = {section: dict(conf.items(section))}
        pprint(section_format)


def get_config(overrides=None):
    conf = ConfigParser()
    files = [DEFAULT_CONFIG]
    user_file = osenv.get('THUG_CONF', None)
    if user_file:
        files.append(user_file)
    conf.read(files)

    if overrides:
        overrides = dict(overrides)
        for section in conf.sections():
            for k in (dict(conf.items(section))):
                if k in overrides:
                    conf[section][k] = overrides.pop(k)

        if overrides:
            LOGGER.warning(
                'Unkown configuration override parameters: {}'.format(
                    ', '.join([str(k) for k, v in overrides.items()])))

    return conf



from unittest.mock import (patch, MagicMock)

from click.testing import CliRunner
import pytest

from thug.cli import meme, thug_meme
from .conftest import IMG_1_FACE


@pytest.fixture
def cli_runner():
    return CliRunner()


@patch('thug.cli.Meme')
class TestMeme:
    def test_it_creates_meme(self, meme_cls, cli_runner):
        res = cli_runner.invoke(
            meme, [IMG_1_FACE, 'my awesome', 'test meme'],
            catch_exceptions=False)

        assert res.exit_code == 0
        assert meme_cls.return_value.create.call_count == 1

        _, meme_kwargs = meme_cls.call_args
        assert meme_kwargs['txt1'] == 'my awesome'
        assert meme_kwargs['txt2'] == 'test meme'


@patch('thug.cli.ThugMeme')
class TestThug:
    @pytest.mark.parametrize('detector_arg, detector',
                             [('opencv', 'thug.cli.HaarCascadeDetector'),
                              ('dlib', 'thug.detect.dlib.DlibDetector')])
    def test_it_creates_thug_meme(self, meme_cls, cli_runner, detector_arg,
                                  detector):
        with patch(detector) as detector:
            detector = detector.return_value
            thugs = MagicMock()
            detector.find_thug_landmarks.return_value = thugs

            res = cli_runner.invoke(
                thug_meme, [
                    IMG_1_FACE, 'my awesome', 'thug meme', '--detector',
                    detector_arg
                ],
                catch_exceptions=False)

            assert res.exit_code == 0
            assert detector.find_thug_landmarks.call_count == 1
            assert meme_cls.return_value.create.call_count == 1

            _, meme_kwargs = meme_cls.call_args
            assert meme_kwargs['thug_landmarks'] == thugs
            assert meme_kwargs['txt1'] == 'my awesome'
            assert meme_kwargs['txt2'] == 'thug meme'

    def test_it_does_not_accept_unknown_detector(self, meme_cls, cli_runner):
        res = cli_runner.invoke(
            thug_meme,
            [IMG_1_FACE, 'my awesome', 'thug meme', '--detector', 'unknown'],
            catch_exceptions=False)
        assert res.exit_code == 2
        assert not meme_cls.return_value.create.call_count


@pytest.mark.parametrize('meme_cls, command',
                         [('thug.cli.Meme', meme),
                          ('thug.cli.ThugMeme', thug_meme)])
class TestCommonCliArgs:
    def test_it_does_not_create_with_invalid_path(self, cli_runner, meme_cls,
                                                  command):
        fpath = 'no-meme-for-you'

        with patch(meme_cls) as meme_cls:
            res = cli_runner.invoke(
                command, [fpath, '', ''], catch_exceptions=False)

            assert res.exit_code == 2
            assert not meme_cls.return_value.create.call_count

    @patch('thug.detect.dlib.DlibDetector')
    @patch('thug.cli.HaarCascadeDetector')
    def test_it_takes_into_account_config_overrides(
            self, detector1, detector2, cli_runner, meme_cls, command):
        o1 = ['-o', 'cigar_length', '0.123']
        o2 = ['-o', 'glasses_width', '0.321']
        o3 = ['-o', 'uknown', 'not going to be used']
        args = [IMG_1_FACE, '', ''] + o1 + o2 + o3

        with patch(meme_cls) as meme_cls:
            res = cli_runner.invoke(command, args, catch_exceptions=False)

            assert res.exit_code == 0

            _, kwargs = meme_cls.call_args
            conf = kwargs['config']
            assert conf['cigar_length'] == '0.123'
            assert conf['glasses_width'] == '0.321'
            assert 'unknown' not in conf



from collections import namedtuple
from math import (atan2, degrees, sqrt, hypot)

Point = namedtuple('Point', 'x y')


class ThugLandmarks:
    def __init__(self, l_eye, r_eye, mouth=None):
        self._l_eye = self._ensure_point_or_none(l_eye)
        self._r_eye = self._ensure_point_or_none(r_eye)
        self._mouth = self._ensure_point_or_none(mouth)
        # require both eyes (I know, a bit discriminatory)
        if self._l_eye:
            assert self._r_eye
        if self._r_eye:
            assert self._l_eye

    @property
    def eyes_available(self):
        return self.left_eye and self.right_eye

    @property
    def mouth_available(self):
        return self.mouth is not None

    @property
    def left_eye(self):
        return self._l_eye

    @property
    def right_eye(self):
        return self._r_eye

    @property
    def eyes_distance(self):
        return hypot(self.eyes_x_difference, self.eyes_y_difference)

    @property
    def eyes_x_difference(self):
        return self.right_eye.x - self.left_eye.x

    @property
    def eyes_y_difference(self):
        return self.right_eye.y - self.left_eye.y

    @property
    def eyes_center(self):
        x = self.left_eye.x + self.eyes_x_difference / 2
        y = self.left_eye.y + self.eyes_y_difference / 2
        return Point(x, y)

    @property
    def mouth(self):
        return self._mouth

    @property
    def eyes_angle(self):
        return degrees(atan2(self.eyes_y_difference, self.eyes_x_difference))

    @staticmethod
    def _ensure_point_or_none(obj):
        if not obj:
            return None
        if not isinstance(obj, Point):
            if not len(obj) == 2:
                raise ValueError('Can not convert to Point: {}'.format(obj))
            return Point(obj[0], obj[1])
        return obj



import cv2
import numpy as np

from .abc import Detector
from .debug import timestamp
from .result import ThugLandmarks
from . import (FACE_CASCADE_FILE, EYE_CASCADE_FILE, MOUTH_CASCADE_FILE)


class HaarCascadeDetector(Detector):
    def __init__(self, config, *args, **kwargs):
        self._face_cascade = cv2.CascadeClassifier(FACE_CASCADE_FILE)
        self._eye_cascade = cv2.CascadeClassifier(EYE_CASCADE_FILE)
        self._mouth_cascade = cv2.CascadeClassifier(MOUTH_CASCADE_FILE)

        self._min_mouth_w = float(config['opencv_min_mouth_w'])
        self._max_mouth_w = float(config['opencv_max_mouth_w'])
        self._min_mouth_h = float(config['opencv_min_mouth_h'])
        self._max_mouth_h = float(config['opencv_max_mouth_h'])

        self._face_scale_factor = float(config['opencv_face_scale_factor'])
        self._face_min_neighbors = int(config['opencv_face_min_neighbors'])
        self._eye_scale_factor = float(config['opencv_eye_scale_factor'])
        self._eye_min_neighbors = int(config['opencv_eye_min_neighbors'])
        self._mouth_scale_factor = float(config['opencv_mouth_scale_factor'])
        self._mouth_min_neighbors = int(config['opencv_mouth_min_neighbors'])

        self._min_eye_w = float(config['opencv_min_eye_w'])
        self._max_eye_w = float(config['opencv_max_eye_w'])
        self._min_eye_h = float(config['opencv_min_eye_h'])
        self._max_eye_h = float(config['opencv_max_eye_h'])

        self._fallback_mouth_x = float(config['opencv_fallback_mouth_x'])
        self._fallback_mouth_y = float(config['opencv_fallback_mouth_y'])

        self._fallback_left_eye_x = float(config['opencv_fallback_left_eye_x'])
        self._fallback_left_eye_y = float(config['opencv_fallback_left_eye_y'])
        self._fallback_right_eye_x = float(
            config['opencv_fallback_right_eye_x'])
        self._fallback_right_eye_y = float(
            config['opencv_fallback_right_eye_y'])

        self._mouth_y_offset = float(config['opencv_mouth_y_offset'])

        super().__init__(config, *args, **kwargs)

    def find_thug_landmarks(self, img_path, show_result=False):
        def to_base(rectangle, base):
            rectangle[0] += base[0]
            rectangle[1] += base[1]
            return rectangle

        img = cv2.imread(img_path)
        gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        faces = self._find_faces(gray)
        faces = self._remove_rects_inside_rects(faces)

        thugs = []
        for face in faces:
            x, y, w, h = face
            base = x, y
            roi_gray = gray[y:y + h, x:x + w]

            eyes, mouth = (None, None), None
            eyes_in_roi, mouth_in_roi = self._find_land_marks(roi_gray)

            if self._values_exist(eyes_in_roi):
                eyes = to_base(eyes_in_roi[0], base), to_base(
                    eyes_in_roi[1], base)

            if self._value_exists(mouth_in_roi):
                mouth = to_base(mouth_in_roi, base)

            thug_result = self._create_thug_detection_result(face, eyes, mouth)

            thugs.append(thug_result)

            if show_result:
                self._draw_result(img, face, eyes, mouth, thug_result)

        if show_result:
            cv2.imshow('detection_result-{}'.format(timestamp()), img)
            cv2.waitKey(1)

        return thugs

    def _find_faces(self, img):
        return self._face_cascade.detectMultiScale(
            img,
            scaleFactor=self._face_scale_factor,
            minNeighbors=self._face_min_neighbors)

    def _find_land_marks(self, face_img_gray):
        eyes = self._find_eyes(face_img_gray)
        mouths = self._find_mouths(face_img_gray)
        return eyes, mouths

    def _find_eyes(self, gray_face_img):
        def find_eye(img):
            results, neighbours, _ = self._eye_cascade.detectMultiScale3(
                img,
                scaleFactor=self._eye_scale_factor,
                minNeighbors=self._eye_min_neighbors,
                minSize=min_size,
                maxSize=max_size,
                outputRejectLevels=True)

            if not self._value_exists(results):
                return None

            eye = results[np.argmax(neighbours)]
            return eye

        h, w = gray_face_img.shape
        min_size = (int(self._min_eye_w * w), int(self._min_eye_h * h))
        max_size = (int(self._max_eye_w * w), int(self._max_eye_h * h))

        half_w = int(w / 2)
        upper_h = int(0.6 * h)
        left_eye_search_area = gray_face_img[:upper_h, :half_w]
        right_eye_search_area = gray_face_img[:upper_h, half_w:]

        left = find_eye(left_eye_search_area)
        right = find_eye(right_eye_search_area)
        if self._value_exists(right):
            right[0] += half_w

        return left, right

    def _find_mouths(self, gray_face_img):
        h, w = gray_face_img.shape
        min_size = (int(self._min_mouth_w * w), int(self._min_mouth_h * h))
        max_size = (int(self._max_mouth_w * w), int(self._max_mouth_h * h))
        half_y = int(h / 2)
        lower_half = gray_face_img[half_y:h]
        results, neighbours, _ = self._mouth_cascade.detectMultiScale3(
            lower_half,
            scaleFactor=self._mouth_scale_factor,
            minNeighbors=self._mouth_min_neighbors,
            minSize=min_size,
            maxSize=max_size,
            outputRejectLevels=True)

        if not self._value_exists(results):
            return None

        best_one = results[np.argmax(neighbours)]
        best_one[1] += half_y

        return best_one

    def _draw_result(self, img, face, eyes, mouth, thug_result):
        self._draw_face(img, face)
        if self._values_exist(eyes):
            self._draw_eyes(img, eyes)
        if self._value_exists(mouth):
            self._draw_mouth(img, mouth)
        self._draw_thug_result(img, thug_result)

    def _draw_eyes(self, img, eyes):
        bgr = (0, 255, 0)
        for eye in eyes:
            self._draw_rectangle(img, eye, bgr)

    def _draw_mouth(self, img, mouth):
        bgr = (0, 0, 255)
        self._draw_rectangle(img, mouth, bgr)

    def _create_thug_detection_result(self, face, eyes, mouth):
        if self._values_exist(eyes):
            # TODO: indexing could be avoided with named tuples
            left_eye = self._calculate_rectange_center(eyes[0])
            right_eye = self._calculate_rectange_center(eyes[1])
        else:
            left_eye, right_eye = self._fallback_thug_result_eyes(face)

        if self._value_exists(mouth):
            mouth = self._calculate_rectange_center(mouth)
            *_, h = face
            mouth = list(mouth)
            mouth[1] += int(self._mouth_y_offset * h)
        else:
            mouth = self._fallback_thug_result_mouth(face)

        return ThugLandmarks(l_eye=left_eye, r_eye=right_eye, mouth=mouth)

    def _fallback_thug_result_eyes(self, face):
        x, y, w, h = face
        left_x = int(self._fallback_left_eye_x * w) + x
        left_y = int(self._fallback_left_eye_y * h) + y
        right_x = int(self._fallback_right_eye_x * w) + x
        right_y = int(self._fallback_right_eye_y * h) + y
        return (left_x, left_y), (right_x, right_y)

    def _fallback_thug_result_mouth(self, face):
        x, y, w, h = face
        mouth_x = int(self._fallback_mouth_x * w) + x
        mouth_y = int(self._fallback_mouth_y * h) + y
        return mouth_x, mouth_y

    @staticmethod
    def _calculate_rectange_center(rect):
        x, y, w, h = rect
        return int(x + w / 2), int(y + h / 2)

    @staticmethod
    def _value_exists(val):
        if isinstance(val, np.ndarray):
            return val.any()
        return val

    def _values_exist(self, vals):
        for val in vals:
            if not self._value_exists(val):
                return False
        return True

    @staticmethod
    def _remove_rects_inside_rects(rects):
        def intersection(a, b):
            x = max(a[0], b[0])
            y = max(a[1], b[1])
            w = min(a[0] + a[2], b[0] + b[2]) - x
            h = min(a[1] + a[3], b[1] + b[3]) - y
            if w < 0 or h < 0:
                return ()
            return (x, y, w, h)

        to_remove = set()
        for idx, r in enumerate(rects):
            *_, w, h = r
            r_area = w * h
            for r_other in rects:
                if not np.array_equal(r, r_other):
                    intersect = intersection(r, r_other)
                    if intersect:
                        *_, wi, hi = intersect
                        intersect_area = wi * hi
                        if intersect_area == r_area:
                            to_remove.add(idx)
        if to_remove:
            rects = np.delete(rects, list(to_remove), 0)

        return rects



import json

import cv2


class Meme:
    def __init__(self, *, config, img_path, txt1, txt2):
        self._text_width_coverage = float(config['text_width_coverage'])
        self._img_path = img_path
        self._txt1 = txt1
        self._txt2 = txt2
        self._font_color = tuple(json.loads(config['font_bgr']))
        self._tmp_result = None
        self._font = int(config['opencv_font'])
        self._font_thickness = int(config['opencv_font_thickness'])
        self._txt_y_offset = float(config['text_y_offset'])
        self._line_type = int(config['opencv_font_line_type'])

    def create(self, res_file, show=True):
        img = cv2.imread(self._img_path)
        img_h, img_w, _ = img.shape

        if self._txt1 or self._txt2:
            txt = self._get_longest_txt()
            scale = self._select_font_scale(img_w, txt)
            loc1, loc2 = self._get_txt_locations(img_w, img_h, scale)

            if self._txt1:
                cv2.putText(img, self._txt1, loc1, self._font, scale,
                            self._font_color, self._font_thickness,
                            self._line_type)
            if self._txt2:
                cv2.putText(img, self._txt2, loc2, self._font, scale,
                            self._font_color, self._font_thickness,
                            self._line_type)

        cv2.imwrite(res_file, img)

        if show:
            from thug.detect.debug import timestamp
            cv2.imshow('meme-{}'.format(timestamp()), img)
            cv2.waitKey(1)

    def _get_longest_txt(self):
        def txt_width(txt):
            return cv2.getTextSize(txt, self._font, 1,
                                   self._font_thickness)[0][0]

        return max((self._txt1, self._txt2), key=txt_width)

    def _select_font_scale(self, img_w, txt):
        scale = 0.5
        step = 0.05
        (w, _), _ = cv2.getTextSize(txt, self._font, scale,
                                    self._font_thickness)
        while w < img_w * self._text_width_coverage:
            scale += step
            (w, _), _ = cv2.getTextSize(txt, self._font, scale,
                                        self._font_thickness)
        return scale

    def _get_txt_locations(self, img_w, img_h, font_scale):
        (txt1_w, txt1_h), _ = cv2.getTextSize(self._txt1, self._font,
                                              font_scale, self._font_thickness)
        (txt2_w, txt2_h), _ = cv2.getTextSize(self._txt2, self._font,
                                              font_scale, self._font_thickness)
        x1 = (img_w - txt1_w) / 2
        x2 = (img_w - txt2_w) / 2
        y_offset = self._txt_y_offset * img_h
        y1 = y_offset + txt1_h
        y2 = img_h - y_offset
        return (int(x1), int(y1)), (int(x2), int(y2))
